/****************************************************************/
/*     Kinfxdb python module, version 0.2                       */
/*       Copyright (C) 1998 Alexander Kuznetsov                 */
/*               alexan@unicorn.micex.ru                        */
/*               updating 2003 rpd Pap                        */  
/*                     pap@any.hu                               */  
/****************************************************************/

#include "Python.h"
#include <time.h>
#include "dbi.h"
#include "sqltypes.h"
#include "locator.h"

#define MAX_CONNECTIONS 20
#define MAX_CURSORS  50
#define MAX_FIELDS   20
#define BLOB_SEGMENT_SIZE 1024

#define MIN(a,b) ((a<b)?a:b)
#define MAX(a,b) ((a>b)?a:b)

/* #define VERBOSE */

EXEC SQL BEGIN DECLARE SECTION;
static char *currentConnection;
EXEC SQL END DECLARE SECTION;

typedef struct vary {
    short          vary_length;
    char           vary_string [1];
} VARY;


/* exception object raised by this module */
static PyObject *informixError;
static PyObject *informixSQLCODE;

typedef struct
{
    int   busy;
    char  name[19];
    int connected;
    int intransaction;
    int isOnline;
} Connection;

static Connection *connections=NULL;  /* contains descriptors for connections */


typedef struct
{
    int busy;  /* is it free to use   */
    int conn;  /* connection object number */
    char  name[19]; /* informix name of the cursor */
    char  stmtname[19]; /* informix name of the cursor */
    struct sqlda *in_sqlda;    /* input parameters */
    struct sqlda *out_sqlda;   /* output parameters */
    char *in_buffer;    /* buffer for parameters */
    char *out_buffer;   /* buffer for result */
    int lastsql_len;
    char *lastsql;
    int prepared;
    int cursoropened;
    int is_select;
} Cursor;

static Cursor *cursors=NULL;


static void setConnection(int i)
{
    if(currentConnection!=connections[i].name)
    {
        currentConnection=connections[i].name;
        EXEC SQL SET CONNECTION :currentConnection;
    }
}


static void raiseSQLException(const char *where)
{
    EXEC SQL BEGIN DECLARE SECTION;
    int exception_count;
    int msglen;
    char msg[256], m[256+sizeof(sqlca.sqlerrm)];
    int i;
    EXEC SQL END DECLARE SECTION;
    PyObject *o, *p;
    int sql_code=SQLCODE;
    if(where)
    {
        o=PyString_FromString(where);
    }
    else
    {
        o=PyString_FromString("");
    }
    EXEC SQL GET DIAGNOSTICS :exception_count=NUMBER;
    for(i=1; i<=exception_count; i++)
    {
        EXEC SQL GET DIAGNOSTICS EXCEPTION :i :msg=MESSAGE_TEXT, :msglen=MESSAGE_LENGTH;
        msg[msglen]=0;
        sprintf(m, msg, sqlca.sqlerrm);
        PyString_ConcatAndDel(&o, PyString_FromString(m));
    }
    PyErr_SetObject( informixError, Py_BuildValue("(iO)", sql_code, o) );
}

static void raiseException(const char *arg)
{
    PyErr_SetObject( informixError, Py_BuildValue("(is)", 0, arg) );
}

/*
   new_connection
   returns new connection object if no available returns -1
 */
static int new_connection()
{
    int i;
    for(i=0; i<MAX_CONNECTIONS; i++)
    {
        if(!connections[i].busy)
        {
            connections[i].busy=1;
            return i;
        }
    }
    return -1;
}

static PyObject *attach_db(PyObject *self, PyObject *args)
{
EXEC SQL BEGIN DECLARE SECTION;
    char *dbname;
    char *user_name=NULL;
    char *user_passwd=NULL;
    char *connectionName;
EXEC SQL END DECLARE SECTION;
    PyObject *o=NULL;
    int conn;
    if((conn=new_connection())<0)
    {
        raiseException("Already reached max number of connections");
        return NULL;
    }
    connectionName=connections[conn].name;
    if(!PyArg_ParseTuple(args, "s|zz",
                         &dbname,
                         &user_name,
                         &user_passwd))
    {
        return NULL;
    }
    if(user_name && user_passwd)
    {
        EXEC SQL CONNECT TO :dbname AS :connectionName USER :user_name USING :user_passwd WITH CONCURRENT TRANSACTION;
    }
    else
    {
        if(!user_passwd)
        {
            EXEC SQL CONNECT TO :dbname AS :connectionName USER :user_name WITH CONCURRENT TRANSACTION;
        }
        else
        {
            EXEC SQL CONNECT TO :dbname AS :connectionName WITH CONCURRENT TRANSACTION;
        }
    }
    if(SQLCODE)
    {
        raiseSQLException("CONNECT: ");
        return NULL;
    }
    connections[conn].connected=1;
    connections[conn].isOnline=sqlca.sqlwarn.sqlwarn3=='W';
    return Py_BuildValue("(ii)", conn, connections[conn].isOnline);
}

static int  start_transaction(int i)
{
    if(connections[i].intransaction)
    {
        return 0;
    }
    EXEC SQL BEGIN WORK;
#ifdef VERBOSE
    printf("BEGIN WORK\n");
#endif
    if(SQLCODE)
    {
        raiseSQLException("BEGIN WORK: ");
        return -1;
    }
    connections[i].intransaction=1;
    return 0;
}

static int  commit_transaction(int i)
{
    if(!connections[i].intransaction)
    {
        return 0;
    }
    EXEC SQL COMMIT WORK;
#ifdef VERBOSE
    printf("COMMIT WORK\n");
#endif
    if(SQLCODE)
    {
        raiseSQLException("COMMIT WORK: ");
        return -1;
    }
    connections[i].intransaction=0;
    return 0;
}

static int  rollback_transaction(int i)
{
    if(!connections[i].intransaction)
    {
        return 0;
    }
    EXEC SQL ROLLBACK WORK;
#ifdef VERBOSE
    printf("ROLLBACK WORK\n");
#endif
    if(SQLCODE)
    {
        raiseSQLException("ROLLBACK WORK: ");
        return -1;
    }
    connections[i].intransaction=0;
    return 0;
}

static void delete_connection(int i)
{
EXEC SQL BEGIN DECLARE SECTION;
    char *connectionName;
EXEC SQL END DECLARE SECTION;
/* also roollback a transaction associated */
    if(!connections[i].busy)
        return;
    connections[i].busy=0;
    if(connections[i].connected)
    {
        if(connections[i].intransaction)
        {
            if(!PyErr_Occurred())
            {
                commit_transaction(i);
            }
            else
            {
                rollback_transaction(i);
            }
        }
    	connectionName=connections[i].name;
        EXEC SQL DISCONNECT :connectionName;
        connections[i].connected=0;
#ifdef VERBOSE
        printf("DISCONNECT %s : %d\n", connectionName, SQLCODE);
#endif
    }
}


static PyObject *detach_db(PyObject *self, PyObject *args)
{
    int conn;
    if(!PyArg_ParseTuple(args, "i", &conn))
    {
        return NULL;
    }
    delete_connection(conn);
    Py_INCREF(Py_None);
    return Py_None;
}


static PyObject *begin(PyObject *self, PyObject *args)
{
    int conn;
    if(!PyArg_ParseTuple(args, "i", &conn))
    {
        return NULL;
    }
    setConnection(conn);
    if(start_transaction(conn)<0)
    {
        return NULL;
    }
    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject *commit(PyObject *self, PyObject *args)
{
    int conn;
    if(!PyArg_ParseTuple(args, "i", &conn))
    {
        return NULL;
    }
    setConnection(conn);
    if(commit_transaction(conn)<0)
    {
        return NULL;
    }
    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject *rollback(PyObject *self, PyObject *args)
{
    int conn;
    if(!PyArg_ParseTuple(args, "i", &conn))
    {
        return NULL;
    }
    setConnection(conn);
    if(rollback_transaction(conn)<0)
    {
        return NULL;
    }
    Py_INCREF(Py_None);
    return Py_None;
}


/*****************************/
/*   CURSOR handle functions */
/*****************************/

static int new_cursor(int conn)
{
    int i;
    for(i=0; i<MAX_CURSORS; i++)
    {
        if(!cursors[i].busy)
        {
            cursors[i].busy=1;
            cursors[i].conn=conn;
            return i;
        }
    }
    raiseException("Max number of cursors reached");
    return -1;
}

static void free_lastsql(int i)
{
    if(cursors[i].lastsql)
    {
        free(cursors[i].lastsql);
        cursors[i].lastsql=NULL;
    }
    cursors[i].lastsql_len=0;
}

static void free_cursor(cur)
{
    EXEC SQL BEGIN DECLARE SECTION;
    char *cursorname;
    EXEC SQL END DECLARE SECTION;
    if(cursors[cur].cursoropened)
    {
        /* Simply close the cursor for statement
         but don't delete it */
        cursorname=cursors[cur].name;
        EXEC SQL CLOSE :cursorname;
        EXEC SQL FREE :cursorname;
        cursors[cur].cursoropened=0;
#ifdef VERBOSE
        printf("FREE %s : %d\n", cursorname, SQLCODE);
#endif
    }
}

static void free_stmt(cur)
{
    EXEC SQL BEGIN DECLARE SECTION;
    char *stmtname;
    EXEC SQL END DECLARE SECTION;
    if(cursors[cur].prepared)
    {
        /* close statement
         but don't delete it */
        stmtname=cursors[cur].stmtname;
        cursors[cur].prepared=0;
        EXEC SQL FREE :stmtname;
#ifdef VERBOSE
        printf("FREE %s : %d\n", stmtname, SQLCODE);
#endif
    }
}

static void free_sqlda(struct sqlda *da)
{
    int i=0;
    struct sqlvar_struct *var;
    loc_t *loc;
    if(da)
    {
        if(da->sqlvar)
        {
            for(i=0; i<da->sqld; i++)
            {
                var=da->sqlvar+i;
                if((var->sqltype==SQLBYTES) ||
                   (var->sqltype==SQLTEXT) )
                {
                    loc=(loc_t*)var->sqldata;
                    if((loc->loc_loctype == LOCMEMORY) &&
                       (loc->loc_bufsize!=-1) &&
                       (loc->loc_buffer))
                    {
                        free(loc->loc_buffer);
                        loc->loc_buffer=NULL;
                    }
                }
            }
        }
        free(da);
    }
}

static void close_cursor(int cur)
/* Closes all cursor resources, but doesn't free the descriptor
   They can be respawn in execute
*/
{
    free_cursor(cur);
    free_stmt(cur);
    if(cursors[cur].in_sqlda->sqlvar)
    {
        free(cursors[cur].in_sqlda->sqlvar);
        cursors[cur].in_sqlda->sqlvar=NULL;
    }
    free_sqlda(cursors[cur].out_sqlda);
    cursors[cur].out_sqlda=NULL;
    if(cursors[cur].in_buffer)
    {
        free(cursors[cur].in_buffer);
        cursors[cur].in_buffer=NULL;
    }
    if(cursors[cur].out_buffer)
    {
        free(cursors[cur].out_buffer);
        cursors[cur].out_buffer=NULL;
    }
    free_lastsql(cur);
}

static PyObject *cursor(PyObject *self, PyObject *args)
{
    int conn;
    int cur;
    if(!PyArg_ParseTuple(args, "i", &conn))
    {
    }
    cur=new_cursor(conn);
    if(cur<0)
    {
        return NULL;
    }
    return PyInt_FromLong(cur);
}

static PyObject *delete_cursor(PyObject *self, PyObject *args)
{
    int cur;
    if(!PyArg_ParseTuple(args, "i", &cur))
    {
        return NULL;
    }
    setConnection(cursors[cur].conn);
    close_cursor(cur);
    cursors[cur].busy=0;     /* reseting to default */
    cursors[cur].conn=-1;
    Py_INCREF(Py_None);
    return Py_None;
}

static int prepare(int cur, char *sql)
{
    EXEC SQL BEGIN DECLARE SECTION;
    char *stmtname;
    char *_sql;
    EXEC SQL END DECLARE SECTION;
    free_stmt(cur);
    stmtname=cursors[cur].stmtname;
    _sql=sql;
    EXEC SQL PREPARE :stmtname FROM :_sql;
    if(SQLCODE)
    {
        raiseSQLException("PREPARE: ");
        return -1;
    }
    cursors[cur].prepared=1;
    return 0;
}

#define INITIAL_SIZE 256
#define INCREMENT INITIAL_SIZE

typedef struct {
    char *ptr;
    int size;
} BUFFER;

static int init_buffer(BUFFER *buffer)
{
    buffer->ptr=malloc(INITIAL_SIZE);
    buffer->size=INITIAL_SIZE;
    return (buffer->ptr)?0:-1;
}

static void free_buffer(BUFFER *buffer)
{
    if(buffer->ptr)
    {
        free(buffer->ptr);
        buffer->ptr=NULL;
    }
}

static int expand_buffer(BUFFER *buffer, int newsize)
{
    
    if(buffer->size < newsize)
    {
        buffer->ptr=realloc(buffer->ptr, newsize+INCREMENT);
        if(!buffer->ptr)
        {
            return -1;
        }
        buffer->size=newsize+INCREMENT;
    }
    return 0;
}


static int set_sqlvar(struct sqlvar_struct *var, BUFFER *buffer, int *shift)
{
    *shift=rtypalign(*shift, var->sqltype);
    if(expand_buffer(buffer, (*shift)+var->sqllen)<0)
    {
        free_buffer(buffer);
        raiseException("expand_buffer: Memory low");
        return -1;
    }
    /* store shift instead of pointers temporary */
    var->sqldata=(char*)*shift;
    return 0;
}

static short FULL_INDICATOR=0;
static short NULL_INDICATOR=-1;

static int bind_input(int cur, PyObject *params)
/*

*/
{
    EXEC SQL BEGIN DECLARE SECTION;
    DATETIME YEAR TO FRACTION dt;
    loc_t loc;
    EXEC SQL END DECLARE SECTION;
    struct sqlda *in_sqlda;
    struct sqlvar_struct *var;
    int i;
    char buf[80];
    BUFFER buffer;
    int shift;
    PyObject *o;
    time_t t;
    struct tm *lt;
    double d;
    int l, k;
    if(!PyTuple_Check(params))
    {
        raiseException("Query parameters must be tuple");
        return -1;
    }
    in_sqlda=cursors[cur].in_sqlda;
    in_sqlda->sqld=PyTuple_Size(params);
    if(in_sqlda->sqlvar)
    {
        free(in_sqlda->sqlvar);
        in_sqlda->sqlvar=NULL;
    }
    in_sqlda->sqlvar=malloc(sizeof(struct sqlvar_struct)*in_sqlda->sqld);
    memset(in_sqlda->sqlvar, 0, sizeof(struct sqlvar_struct)*in_sqlda->sqld);
    if(cursors[cur].in_buffer)
    {
        free(cursors[cur].in_buffer);
        cursors[cur].in_buffer=NULL;
    }
    shift=0;
    if(init_buffer(&buffer)<0)
    {
        raiseException("init_buffer: Memory low");
        return -1;
    }
    for(i=0; i<in_sqlda->sqld; i++)
    {
        o=PyTuple_GetItem(params, i);
        var=in_sqlda->sqlvar+i;
        var->sqlind=&FULL_INDICATOR;
        if(o==Py_None)
        {
            var->sqltype=CFIXCHARTYPE;
            var->sqllen=0;
            var->sqldata=0;
            var->sqlind=&NULL_INDICATOR;
        }else
        if(PyInt_Check(o))
        {
            var->sqltype=CLONGTYPE;
            var->sqllen=sizeof(long);
            if(set_sqlvar(var, &buffer, &shift)<0) return -1;
            *(int*)(buffer.ptr+shift)=PyInt_AsLong(o);
        }else
        if(PyLong_Check(o))
        {
            var->sqltype=CLONGTYPE;
            var->sqllen=sizeof(long);
            if(set_sqlvar(var, &buffer, &shift)<0) return -1;
            *(long*)(buffer.ptr+shift)=PyInt_AsLong(o);
        }else
        if(PyString_Check(o))
        {
            var->sqltype=CFIXCHARTYPE;
            var->sqllen=PyString_Size(o);
            if(set_sqlvar(var, &buffer, &shift)<0) return -1;
            memcpy(buffer.ptr+shift, PyString_AsString(o), var->sqllen );
        }else
        if(PyFloat_Check(o))
        {
            var->sqltype=CDOUBLETYPE;
            var->sqllen=sizeof(double);
            if(set_sqlvar(var, &buffer, &shift)<0) return -1;
            *(double*)(buffer.ptr+shift)=PyFloat_AsDouble(o);
        }else
        if(dbiIsRaw(o))
        {
            var->sqltype=CLOCATORTYPE;
            o=dbiValue(o);
            if(!PyString_Check(o))
            {
                free_buffer(&buffer);
                raiseException("dbiRaw: Value should be string");
                return -1;
            }
            var->sqllen=sizeof(loc_t);
            if(set_sqlvar(var, &buffer, &shift)<0) return -1;
            loc.loc_loctype = LOCMEMORY;
            loc.loc_buffer = PyString_AsString(o);
            loc.loc_bufsize = PyString_Size(o);
            loc.loc_size = loc.loc_bufsize;
            loc.loc_oflags = 0;
            loc.loc_mflags = 0;
            *(loc_t*)(buffer.ptr+shift)=loc;
        }else
        if(dbiIsDate(o))
        {
            var->sqltype=CDTIMETYPE;
            var->sqllen=sizeof(dtime_t);
            if(set_sqlvar(var, &buffer, &shift)<0) return -1;
            o=dbiValue(o);
            if(PyInt_Check(o) || PyLong_Check(o))
            {
                l=PyInt_AsLong(o);
                lt=(struct tm *)localtime((time_t*)&l);
                sprintf(buf, "%02d.%02d.%04d %02d:%02d:%02d.%3d",
                        lt->tm_mday,
                        lt->tm_mon+1,
                        lt->tm_year+1900,
                        lt->tm_hour,
                        lt->tm_min,
                        lt->tm_sec,
                        0
                       );
            }else
            if(PyFloat_Check(o))
            {
                d=PyFloat_AsDouble(o);
                l=(int)d;
                k=(int)((d-l)*1000+0.5);
                if(k>999)
                {
                    k=0;
                }
                lt=(struct tm *)localtime((time_t*)&l);
                sprintf(buf, "%02d.%02d.%04d %02d:%02d:%02d.%3d",
                        lt->tm_mday,
                        lt->tm_mon+1,
                        lt->tm_year+1900,
                        lt->tm_hour,
                        lt->tm_min,
                        lt->tm_sec,
                        k
                       );
            }else
            {
                free_buffer(&buffer);
                raiseException("dbiDate: invalid value type");
                return -1;
            }
        /*    printf("%s\n", buf); */
            if(dtcvfmtasc(buf,"%d.%m.%Y %H:%M:%S%F3", &dt))
            {
                free_buffer(&buffer);
                raiseException("dtcvfmtasc: internal error");
                return -1;
            }
            *(dtime_t*)(buffer.ptr+shift)=dt;
        }else
            {
                free_buffer(&buffer);
                sprintf(buf, "bind_input: parameter %d type not supported", i);
                raiseException(buf);
                return -1;
            }
        shift+=var->sqllen;
    }
    for(i=0; i<in_sqlda->sqld; i++)
    {
        var=in_sqlda->sqlvar+i;
        var->sqldata=buffer.ptr+(int)(var->sqldata);
    }
    cursors[cur].in_buffer=buffer.ptr;
    return 0;
}

static PyObject *typeOf(int type)
{
  switch(type & SQLTYPE){
  case SQLBYTES:
  case SQLTEXT:
    return DbiRaw;
  case SQLDATE:
  case SQLDTIME:
    return DbiDate;
  case SQLFLOAT:
  case SQLSMFLOAT:
  case SQLDECIMAL:
  case SQLINT:
  case SQLSMINT:
  case SQLSERIAL:
  case SQLMONEY:
  case SQLINTERVAL:
    return DbiNumber;
  default:
    return DbiString;
  }
}


static PyObject *description(int cur)
{
    struct sqlvar_struct *var;
    struct sqlda *da=cursors[cur].out_sqlda;
    int i;
    PyObject *o=NULL;
    PyObject *des=NULL;
    PyObject *t=NULL;
    des=PyTuple_New(da->sqld);
    for(i=0; i<da->sqld; i++)
    {
        var=da->sqlvar+i;
        o=typeOf(var->sqltype);
        Py_INCREF(o);
        PyTuple_SetItem(des, i, Py_BuildValue("(sOiiiii)",
					var->sqlname, 
					o,
					var->sqllen,
					var->sqllen,
					0, 0, !(var->sqltype & SQLNONULL)));
    }
    return des;
}

static int allocate_outputbuffer(int cur)
{
    loc_t *loc;
    char *buffer;
    struct sqlvar_struct *var;
    struct sqlda *da=cursors[cur].out_sqlda;
    int shift;
    int i;
    short sqltype;
    int len;
    char buf[80];
    EXEC SQL BEGIN DECLARE SECTION;
    DATETIME YEAR TO FRACTION(5) dt;
    EXEC SQL END DECLARE SECTION;
    
    if(cursors[cur].out_buffer)
    {
        free(cursors[cur].out_buffer);
        cursors[cur].out_buffer=NULL;
    }
    shift=0;
    for(i=0; i<da->sqld; i++)
    {
        var=da->sqlvar+i;
        sqltype=var->sqltype & SQLTYPE;
        len=rtypmsize(sqltype, var->sqllen);
        switch(sqltype)
        {
        case SQLCHAR:
        case SQLVCHAR:
        case SQLNCHAR:
        case SQLNVCHAR:
            len++;
            var->sqllen++;
            break;
        case SQLINT:
        case SQLSMINT:
        case SQLFLOAT:
        case SQLSMFLOAT:
        case SQLDECIMAL:
        case SQLSERIAL:
        case SQLDATE:
        case SQLMONEY:
        case SQLBYTES:
        case SQLTEXT:
        case SQLINTERVAL:
            break;
        case SQLDTIME:
#if PLATFORM != Linux
            len=sizeof(dtime_t);
            var->sqllen=len;
#endif
            break;
        default:
            free(cursors[cur].out_sqlda);
            cursors[cur].out_sqlda=NULL;
            sprintf(buf, "Type not supported %s", rtypname(sqltype));
            raiseException(buf);
            return -1;
            
        }
        shift=rtypalign(shift, var->sqltype);
        var->sqldata=(char*)shift;
        shift+=len;
        shift=rtypalign(shift, CSHORTTYPE);
        var->sqlind=(short*)shift;
        shift+=sizeof(short);
    }
    buffer=(char *)malloc(shift);
    if(!buffer)
    {
        raiseException("allocate_outputbuffer: Memory low");
        return -1;
    }
    
    for(i=0; i<da->sqld; i++)
    {
        var=da->sqlvar+i;
        /* setting up a real pointers */
        var->sqldata=buffer+(int)(var->sqldata);
        var->sqlind=(short *)(buffer+(int)(var->sqlind));
        sqltype=var->sqltype & SQLTYPE;
        switch(sqltype)
        {
        case SQLBYTES:
        case SQLTEXT:
            loc=(loc_t*)var->sqldata;
            loc->loc_loctype = LOCMEMORY;
            loc->loc_bufsize = -1;
            loc->loc_oflags = 0;
            loc->loc_mflags = 0;
            loc->loc_buffer = NULL;
            break;
        case SQLDATE:
            *(dtime_t*)(var->sqldata)=dt;
        }
    }
    cursors[cur].out_buffer=buffer;
    return 0;
}

static PyObject *execute(PyObject *self, PyObject *args)
{
    EXEC SQL BEGIN DECLARE SECTION;
    char *stmtname;
    char *cursorname;
    struct sqlda *in_sqlda=NULL;
    struct sqlda *out_sqlda=NULL;
    EXEC SQL END DECLARE SECTION;
    int cur;
    int conn;
    char *sql;
    int sqllen;
    PyObject *params=NULL;
    PyObject *o;
    PyObject *descr=NULL;
    int descr_again;
    int need_prep;
    if(!PyArg_ParseTuple(args, "is#|O!", &cur, &sql, &sqllen, &PyTuple_Type, &params))
    {
        return NULL;
    }
    conn=cursors[cur].conn;
    setConnection(conn);
    free_cursor(cur);
    need_prep=1;
    if(cursors[cur].prepared && cursors[cur].lastsql)
        if(!strncmp(sql, cursors[cur].lastsql, MIN(sqllen, cursors[cur].lastsql_len)))
        {
            need_prep=0;
        }
    if(need_prep)
    {
        if(prepare(cur, sql)<0)
        {
            return NULL;
        }
        stmtname=cursors[cur].stmtname;
        EXEC SQL DESCRIBE :stmtname INTO out_sqlda;
        if(SQLCODE<0)
        {
            free(out_sqlda);
            out_sqlda=NULL;
            raiseSQLException("DESCRIBE: ");
            return NULL;
        }else
        if(SQLCODE==0||SQLCODE==56) 
        { /* SQLCODE==0 - select statement SQLCODE==56 - execute procedure statement */    
	    if(out_sqlda->sqld>0) 
	    {/*if description count greater zero then return smth */
    		cursors[cur].is_select=1;
    		free_sqlda(cursors[cur].out_sqlda);
    		cursors[cur].out_sqlda=out_sqlda;
    		if(allocate_outputbuffer(cur)<0)
    		{
    		    return NULL;
    		}
    		descr=description(cur);
	    }
	    else 
	    { /* if does not descriptions then does not return values */
        	cursors[cur].is_select=0;
        	if(out_sqlda)
        	{
            	  free(out_sqlda);
        	  out_sqlda=NULL;
        	}
	    }
	}else
        {  /* does not return values */
            cursors[cur].is_select=0;
            if(out_sqlda)
            {
              free(out_sqlda);
              out_sqlda=NULL;
            }
        }
    }
    
    if(params)
    {
        if(bind_input(cur, params)<0)
        {
            Py_XDECREF(descr);
            return NULL;
        }
    }
    stmtname=cursors[cur].stmtname;
    if(cursors[cur].is_select)
    { /*   cursors[cur].cursoropened */
        cursorname=cursors[cur].name;
        EXEC SQL DECLARE :cursorname CURSOR WITH HOLD FOR :stmtname;
        if(SQLCODE)
        {
            raiseSQLException("DECLARE: ");
            Py_XDECREF(descr);
            return NULL;
        }
        cursors[cur].cursoropened=1;
        if(params)
        {
            in_sqlda=cursors[cur].in_sqlda;
            EXEC SQL OPEN :cursorname USING DESCRIPTOR in_sqlda;
        }
        else
        {
            EXEC SQL OPEN :cursorname;
        }
        if(SQLCODE)
        {
            raiseSQLException("OPEN: ");
            Py_XDECREF(descr);
            return NULL;
        }
    }
    else
    {
        if(params)
        {
            in_sqlda=cursors[cur].in_sqlda;
            EXEC SQL EXECUTE :stmtname USING DESCRIPTOR in_sqlda;
        }
        else
        {
            EXEC SQL EXECUTE :stmtname;
        }
        if(SQLCODE)
        {
            raiseSQLException("EXECUTE: ");
            Py_XDECREF(descr);
            return NULL;
        }
    }
    if(need_prep)
    {
        if(cursors[cur].lastsql)
        {
            free(cursors[cur].lastsql);
            cursors[cur].lastsql=NULL;
        }
        cursors[cur].lastsql=(char*)malloc(sqllen+1);
        memcpy(cursors[cur].lastsql, sql, sqllen);
        cursors[cur].lastsql_len=sqllen;
    }
    if(descr)
    {
        o=Py_BuildValue("(OO)",Py_None, descr);
        Py_DECREF(descr);
    }
    else
    {
        o=Py_BuildValue("(iO)", 1, Py_None);
    }
    return o;
}


static PyObject *sqlvar2PyObject(struct sqlvar_struct *var)
{
    PyObject *o;
    short sqltype;
    char buf[80];
    double d;
    struct tm lt;
    time_t t;
    int frac;
    short mdy[3];

    sqltype=var->sqltype & SQLTYPE;
    if(*var->sqlind == -1)
    {
        Py_INCREF(Py_None);
        return Py_None;
    }
    switch(sqltype)
    {
    case SQLCHAR:
    case SQLVCHAR:
    case SQLNCHAR:
    case SQLNVCHAR:
 /*       printf("sqldata: %s, var->sqllen: %d", var->sqldata, var->sqllen); */
        o=PyString_FromString(var->sqldata);
        break;
    case SQLFLOAT:
        o=PyFloat_FromDouble(*(double*)(var->sqldata));
        break;
    case SQLSMFLOAT:
        o=PyFloat_FromDouble((double)*(float*)(var->sqldata));
        break;
    case SQLDECIMAL:
    case SQLMONEY:
        dectodbl((dec_t*)var->sqldata, &d);
        o=PyFloat_FromDouble(d);
        break;
    case SQLINT:
    case SQLSERIAL:
        o=PyInt_FromLong((long)*(int*)(var->sqldata));
        break;
    case SQLSMINT:
        o=PyInt_FromLong((long)*(short*)(var->sqldata));
        break;
    case SQLBYTES:
    case SQLTEXT:
        o=dbiMakeRaw(PyString_FromStringAndSize(((loc_t*)(var->sqldata))->loc_buffer,
                                     ((loc_t*)(var->sqldata))->loc_size));
        break;
    case SQLDTIME:
        if(dttofmtasc((dtime_t*)(var->sqldata), buf, sizeof(buf), "%d.%m.%Y %H:%M:%S%F5"))
        {
            raiseException("internal error: dttofmtasc failed");
            return NULL;
        }
        sscanf(buf, "%02d.%02d.%04d %02d:%02d:%02d.%05d",
               &lt.tm_mday,
               &lt.tm_mon,
               &lt.tm_year,
               &lt.tm_hour,
               &lt.tm_min,
               &lt.tm_sec,
               &frac);
        lt.tm_year -= 1900;
        lt.tm_mon -= 1;
        lt.tm_isdst = -1;
        d=((double)frac)/100000;
        o=dbiMakeDate(PyFloat_FromDouble(mktime(&lt)+d));
        break;
    case SQLDATE:
        rjulmdy(*(int *)(var->sqldata), mdy);
        lt.tm_year=mdy[2]-1900;
        lt.tm_mon=mdy[0]-1;
        lt.tm_mday=mdy[1];
        lt.tm_hour=lt.tm_min=lt.tm_sec=0;
        lt.tm_isdst = -1;
        o=dbiMakeDate(PyFloat_FromDouble((double)mktime(&lt)));
        break;
        
    case SQLINTERVAL:
        if(intofmtasc((intrvl_t*)(var->sqldata), buf, sizeof(buf), "%d.%m.%Y %H:%M:%S%F5"))
        {
            raiseException("internal error:intofmtasc failed");
            return NULL;
        }
        sscanf(buf, "%02d.%02d.%04d %02d:%02d:%02d.%05d",
               &lt.tm_mday,
               &lt.tm_mon,
               &lt.tm_year,
               &lt.tm_hour,
               &lt.tm_min,
               &lt.tm_sec,
               &frac);
        lt.tm_year -= 1900;
        lt.tm_mon -= 1;
        lt.tm_isdst = -1;
        d=((double)frac)/100000;
        o=PyFloat_FromDouble(mktime(&lt)+d);
        break;
 
    default:
        sprintf(buf, "Type %s(%d) not supported in fetch", rtypname(sqltype), sqltype);
        raiseException(buf);
        return NULL;
    }
    return o;
}

static PyObject *sqlda2Tuple(struct sqlda *da)
{
    PyObject *res;
    PyObject *o;
    int i;
    res=PyTuple_New(da->sqld);
    for(i=0; i<da->sqld; i++)
    {
        o=sqlvar2PyObject(da->sqlvar+i);
        if(!o)
        {
            Py_DECREF(res);
            return NULL;
        }
        PyTuple_SetItem(res, i, o);
    }
    return res;
}

static PyObject *fetch(PyObject *self, PyObject *args)
{
    EXEC SQL BEGIN DECLARE SECTION;
    char *cursorName;
    struct sqlda *out_sqlda;
    EXEC SQL END DECLARE SECTION;
    int cur;
    int conn;
    int fetch_stat;
    PyObject *o;
    if(!PyArg_ParseTuple(args, "i", &cur))
    {
        return NULL;
    }
    if(!cursors[cur].cursoropened)
    {
        raiseException("fetch: cursor not opened");
        close_cursor(cur);
        return NULL;
    }
    conn=cursors[cur].conn;
    setConnection(conn);
    cursorName=cursors[cur].name;
    out_sqlda=cursors[cur].out_sqlda;
    EXEC SQL FETCH :cursorName USING DESCRIPTOR out_sqlda;
    switch(SQLCODE)
    {
    case 100:
        /* end of records */
        free_cursor(cur);
        Py_INCREF(Py_None);
        return Py_None;
    case 0:
        /* succesfull fetch  */
        o=sqlda2Tuple(cursors[cur].out_sqlda);
        if(!o)
        {
            close_cursor(cur);
            return NULL;
        }
        break;
    default:
        raiseSQLException("FETCH: ");
        close_cursor(cur);
        return NULL;
    }
    return o;
}

static PyObject *_close_cursor(PyObject *self, PyObject *args)
{
    int cur;
    int conn;
    if(!PyArg_ParseTuple(args, "i", &cur))
    {
        return NULL;
    }
    conn=cursors[cur].conn;
    setConnection(conn);
    close_cursor(cur);
    Py_INCREF(Py_None);
    return Py_None;
}

/* Global methods declaration */

static PyMethodDef kinfxdb_GlobalMethods[] =
{
    {"attach_db",  attach_db, 1},
    {"detach_db",  detach_db, 1},
    {"begin",  begin, 1},
    {"commit", commit, 1},
    {"rollback", rollback, 1},
    {"cursor", cursor, 1},
    {"free_cursor", delete_cursor, 1},
    {"execute", execute, 1},
    {"fetch", fetch, 1},
    {"close", _close_cursor, 1},
    {NULL,      NULL}       
};

void init_kinfxdb()
{
    
    PyObject *m, *d;
    int i;
    char buf[20];
    m = Py_InitModule("_kinfxdb", kinfxdb_GlobalMethods);
    d = PyModule_GetDict(m);
    informixError = PyString_FromString("kinfxdb.error");
    informixSQLCODE = PyLong_FromLong(0L);
    PyDict_SetItemString(d, "error", informixError);
    PyDict_SetItemString(d, "SQLCODE", informixSQLCODE);
    connections=malloc(sizeof(Connection)*MAX_CONNECTIONS);
    for(i=0; i<MAX_CONNECTIONS; i++)
    {
        connections[i].busy=0;
        sprintf(connections[i].name, "conn%d", i);
        connections[i].connected=0;
        connections[i].intransaction=0;
    }                   
    cursors=malloc(sizeof(Cursor)*MAX_CURSORS);
    for(i=0; i<MAX_CURSORS; i++)
    {
        cursors[i].busy=0;
        cursors[i].conn=-1;
        cursors[i].in_sqlda=malloc(sizeof(struct sqlda));
        memset(cursors[i].in_sqlda, 0, sizeof(struct sqlda));
        cursors[i].out_sqlda=NULL;
        cursors[i].in_buffer=NULL;
        cursors[i].out_buffer=NULL;
        cursors[i].lastsql=NULL;
        cursors[i].lastsql_len=0;
        sprintf(cursors[i].name, "cursor%d", i);
        sprintf(cursors[i].stmtname, "statement%d", i);
        cursors[i].prepared=0;
        cursors[i].cursoropened=0;
    }
    initdbi(d);
}

